from .metric import *
from .utils import *

from .text_diversities.morphological import *
from .text_diversities.phonological import *
from .text_diversities.syntactic import *
from .text_diversities.semantic import *

name = "textdiversity"


# TextDiversity pkgs
import spacy
import torch
import numpy as np
import itertools
from functools import partial
from sklearn.decomposition import PCA

from Bio import Align

# locals
from ..metric import TextDiversity
from ..utils import *

# ==================================================================================
# helper functions 
# ==================================================================================

def align_and_score(seq1, seq2):
    return Align.PairwiseAligner().align(seq1, seq2).score

# ==================================================================================

class POSSequenceDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'dim_reducer': PCA,
        'remove_stopwords': False, 
        'verbose': False,
        # POSSequenceDiversity configs
        'pad_to_max_len': False, 
        'pos_to_alpha' : True,
        'use_gpu': False,
        'n_components': None 
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)
        self.model = spacy.load("en_core_web_sm")

    def extract_features(self, corpus):

        # clean corpus
        corpus = clean_text(corpus)

        # split sentences
        sentences = split_sentences(corpus)

        # extracts parts-of-speech (poses)
        poses = []
        for s in sentences:
            pos = [token.pos_ for token in self.model(s)] #if token.text not in stopwords]
            poses.append(pos)

        # compute max seq length for padding / normalization later
        self.max_len = find_max_list(poses)

        # pad to max sentence doc length
        if self.config['pad_to_max_len']:
            poses = np.array([s + ['NULL'] * (self.max_len - len(s)) for s in poses])

        # convert poses to alpha
        if self.config['pos_to_alpha']:
            # build dict of unique poses
            pos_map = set(itertools.chain(*poses))
            pos_map = {tag: chr(i+65) for i, tag in enumerate(pos_map)}
            # convert to int for distance comparison
            if isinstance(poses, np.ndarray):
                pos_to_alpha_fn = np.vectorize(pos_map.get)
                poses = pos_to_alpha_fn(poses)
            else:
                poses = [list(map(pos_map.get, pos)) for pos in poses]

        # create strands of morphological dna
        pos_strands = ["".join(pos) for pos in poses]

        return pos_strands, sentences

    def calculate_similarities(self, features):
        """
        Uses biopython.Bio.Align.PairwiseAligner() to align and score stands of
        morphological DNA (i.e. sequences of pos tags)
        """

        # compute pairwise alignment + scoring
        Z = compute_pairwise(
            features, 
            align_and_score, 
            self.max_len, 
            verbose=self.config['verbose'])

        # normalize the similarities by max_len
        Z = Z / self.max_len

        return Z

    def calculate_abundance(self, species):
        num_species = len(species)
        p = np.full(num_species, 1 / num_species)
        return p

    def __call__(self, corpus): 
        return super().__call__(corpus)


if __name__ == '__main__':

    # TEST
    lo_div = ['one massive earth', 'an enormous globe', 'the colossal world']
    hi_div = ['basic human right', 'you were right', 'make a right']

    # diversities
    print("diversities")
    print_div_metric(POSSequenceDiversity, lo_div, hi_div)

    # similarities
    print("similarities")
    print_sim_metric(POSSequenceDiversity, lo_div, hi_div)


from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
import seaborn as sns 
from tqdm import tqdm
import numpy as np
import tempfile
import os
import csv

class Metric(ABC):
    use_me = False  # static var indicates to run files whether or not to use this metric
    default_config = {}  # static var, specifies the default config for run files

    def __init__(self, config):
        self.config = config

        # validate config
        assert type(self.config) == dict, 'Metric config must be dict type.'

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    def uint_assert(self, field_name):
        err_msg = 'Required: {}(int) > 0'.format(field_name)
        assert type(self.config.get(field_name, None)) == int, err_msg
        assert self.config[field_name] > 0, err_msg

    def input_path_assert(self, field_name):
        err_msg = '[{}] not exists.'.format(field_name)
        assert os.path.exists(self.config.get(field_name, None)), err_msg


class DiversityMetric(Metric):
    required_input = 'response_set'  # in most cases, the diversity metric input is the response set S_c

    def __init__(self, config):
        super().__init__(config)

    @abstractmethod
    def __call__(self, response_set):
        # validate input
        assert type(response_set) == list
        assert all([type(e) == str for e in response_set])

        # place holder
        diversity_score = None
        return diversity_score


class SimilarityMetric(Metric):

    def __init__(self, config):
        super().__init__(config)

    @abstractmethod
    def __call__(self, resp_a, resp_b):
        # validate input
        assert type(resp_a) == type(resp_b) == str

        # place holder
        similarity_score = None
        return similarity_score
        

class TextDiversity(DiversityMetric):

    def __init__(self, config):
        super().__init__(config)

    def __call__(self, corpus):

        if not all([type(d) == str and d.strip() != "" for d in corpus]):
            print('corpus contains invalid inputs: \n {} \n returning 0'.format(corpus))
            return 0

        # extract features + species
        features, species = self.extract_features(corpus)

        # if there are no features, diversity is 0 by default
        if len(features) == 0:
            return 0

        # get similarity matrix Z
        Z = self.calculate_similarities(features)
        
        # get abundance vector p
        p = self.calculate_abundance(species)

        # calculate diversity
        D = self.calc_div(p, Z, self.config['q'])

        # optionally normalize diversity by number of species
        # which is the length of the p vector
        if self.config['normalize']:
            D /= len(p)

        return D

    def similarity(self, corpus):

        if not all([type(d) == str and d.strip() != "" for d in corpus]):
            print('corpus contains invalid inputs: \n {} \n returning 0'.format(corpus))
            return 0

        # extract features + species
        features, species = self.extract_features(corpus)

        # if there are no features, diversity is 0 by default
        if len(features) == 0:
            return 0

        # get similarity matrix Z
        Z = self.calculate_similarities(features)

        # calculate similarity score
        similarity = Z.sum() / (len(Z) ** 2)
        
        return similarity

    @abstractmethod
    def extract_features(self, corpus, *args, **kwargs):  
        # validate input
        assert type(corpus) == list
        assert all([type(d) == str for d in corpus])

        # place holders
        features, species = None, None
        return features, species 

    @abstractmethod
    def calculate_similarities(self, features):

        # place holder
        Z = None
        return Z

    @abstractmethod
    def calculate_abundance(self, species):

        # place holder
        p = None
        return p
        
        # num_embeddings = len(embs)

        # Z = np.eye(num_embeddings)
        # iu = np.triu_indices(num_embeddings, k=1)
        # il = (iu[1], iu[0])

        # iterable = range(num_embeddings)
        # if self.verbose:
        #     print('calculating similarity matrix...')
        #     iterable = tqdm(iterable)

        # for e1 in iterable:
        #     for e2 in range(1, num_embeddings - e1):
        #         d = self.config['distance_fn'](embs[e1], embs[e1 + e2])
        #         if self.config['scale_dist'] == "exp":
        #             d = np.exp(-d)
        #         elif self.config['scale_dist'] == "invert":
        #             d = 1 - d
        #         Z[e1][e1 + e2] = d     
        # Z[il] = Z[iu]

        # # remove some noise from the Z similarities
        # if self.config['sq_reg']:
        #     Z **= 2 

        # # remove some noise from the Z similarities
        # if self.config['mean_adj']:
        #     off_diag = np.where(~np.eye(Z.shape[0],dtype=bool))
        #     Z[off_diag] -= Z[off_diag].mean()
        #     Z = np.where(Z < 0, 0 , Z)

        # return Z

    def calc_div(self, p, Z, q=1):
        Zp =  Z @ p
        if q == 1:
            D = 1 / np.prod(Zp ** p)
        elif q == float('inf'):
            D = 1 / Zp.max()
        else:
            D = (p * Zp ** (q-1)).sum() ** (1/(1-q))
        return D  

    def diversity_profile(self, response_set, range=None):

        # embed inputs
        embs, species = self.get_embeddings(response_set)

        if len(embs) == 0:
            return 0

        # get similarity matrix 
        Z = self.calc_Z(embs)

        # get diversities for all q in range
        if range is None:
            range = np.arange(0, 101)

        num_species = len(species)
        p = np.full(num_species, 1/num_species)

        Ds = []
        for q in range:
            D = self.calc_div(p, Z, q)
            Ds.append(D)

        # plot diversity profile
        ax = sns.lineplot(x=range, y=Ds)  
        ax.set(xlabel="Sensitivity Parameter, $q$", 
               ylabel="Diversity $^qD^{\mathbf{Z}}(\mathbf{p})$", 
               title="Corpus Diversity Profile")
        plt.show()

    def species_heatmap(self, response_set, n=10):

        # embed inputs
        embs, species = self.get_embeddings(response_set)

        if len(embs) == 0:
            return 0

        # get similarity matrix 
        Z = self.calc_Z(embs)

        plt.figure(figsize = (10,8))

        g = sns.heatmap(
            data=np.around(Z[:n,:n], 2), 
            xticklabels=species[:n],
            yticklabels=species[:n],
            vmin=0,
            vmax=1,
            annot=True, 
            annot_kws={"fontsize": 10}, 
            fmt='g')
        g.set_xticklabels(g.get_xticklabels(), rotation=90)
        g.set_title('Similarities')
        plt.show()


# TextDiversity pkgs
import spacy
import torch
import numpy as np
import pandas as pd
import itertools
from functools import partial
import logging
logging.basicConfig(level=logging.CRITICAL)

from Bio import Align

# rhythmic diversity
import cadence as cd
import string

# phonemic diversity
from phonemizer import phonemize
from phonemizer.backend import EspeakBackend
from phonemizer.punctuation import Punctuation
from phonemizer.separator import Separator

import os
import sys
from collections import Counter

# locals
from ..metric import TextDiversity
from ..utils import *

# ==================================================================================
# helper functions 
# ==================================================================================

def align_and_score(seq1, seq2):
    return Align.PairwiseAligner().align(seq1, seq2).score

# ==================================================================================

class RhythmicDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'remove_stopwords': False, 
        'verbose': False,
        # RhythmicDiversity configs
        'pad_to_max_len': False, 
        'rhythm_to_alpha' : True
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)
        self.model = spacy.load("en_core_web_sm")

    def extract_features(self, corpus):

        # clean corpus
        corpus = clean_text(corpus)

        # split sentences
        sentences = split_sentences(corpus)

        # strip punctuation
        sentences = [s.translate(str.maketrans('', '', string.punctuation)) for s in sentences]

        # extracts rhythms (sequences of [un]weighted [un]stressed syllables)
        rhythms = []
        for s in sentences:
            prose = cd.Prose(s)
            df = prose.sylls().reset_index()
            if all([c in df.columns for c in ['syll_stress', 'syll_weight']]):
                rhythm = df[df['word_ipa_i'] == 1][['syll_stress', 'syll_weight']].values
                rhythm = ["".join(row) for row in rhythm]
            else:
                rhythm = [""]
            rhythms.append(rhythm)

        # compute max seq length for padding / normalization later
        self.max_len = find_max_list(rhythms)

        # pad to max sentence doc length
        if self.config['pad_to_max_len']:
            rhythms = np.array([r + ['N'] * (self.max_len - len(r)) for r in rhythms])

        # convert rhythms to alpha
        if self.config['rhythm_to_alpha']:
            # build dict of unique rhythms
            rhythm_map = set(itertools.chain(*rhythms))
            rhythm_map = {tag: chr(i+65) for i, tag in enumerate(rhythm_map)}
            # convert to int for distance comparison
            if isinstance(rhythm, np.ndarray):
                rhythm_to_alpha_fn = np.vectorize(rhythm_map.get)
                rhythm = rhythm_to_alpha_fn(rhythm)
            else:
                rhythms = [list(map(rhythm_map.get, r)) for r in rhythms]

        # create strands of rhythmic dna
        rhythm_strands = ["".join(r) for r in rhythms]

        return rhythm_strands, sentences

    def calculate_similarities(self, features):
        """
        Uses biopython.Bio.Align.PairwiseAligner() to align and score stands of
        morphological DNA (i.e. sequences of [un]weighted [un]stressed syllables)
        """

        # compute pairwise alignment + scoring
        Z = compute_pairwise(
            features, 
            align_and_score, 
            self.max_len, 
            verbose=self.config['verbose'])

        # normalize the similarities by max_len
        Z = Z / self.max_len

        return Z

    def calculate_abundance(self, species):
        num_species = len(species)
        p = np.full(num_species, 1 / num_species)
        return p

    def __call__(self, corpus): 
        return super().__call__(corpus)




class PhonemicDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'remove_stopwords': False, 
        'verbose': False,
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)

        # load pre-computed phonemic similarity matrix Z
        file_dir = os.path.dirname(__file__)
        csv_path = os.path.join(file_dir, '../similarities/phoneme_similarities.csv')
        self.Z = pd.read_csv(csv_path, index_col=0)

        # phonemizer
        logger = logging.getLogger()
        logger.disabled = True
        self.backend = EspeakBackend('en-us', words_mismatch='ignore', logger=logger)
        self.separator = Separator(phone=' ', word=None)

    def get_phonemes(self, corpus, return_counts=True):
        # remove all the punctuation from the text, considering only the specified
        # punctuation marks
        text = Punctuation(';:,.!"?()-/><').remove(corpus)

        # build the set of all the words in the text
        words = {w.lower() for line in text for w in line.strip().split(' ') if w}

        # build the lexicon by phonemizing each word one by one. The backend.phonemize
        # function expect a list as input and outputs a list.
        lexicon = {
            word : self.backend.phonemize([word], separator=self.separator, strip=True)[0]
            for word in words} 

        if return_counts:
            phoneme_counts = Counter(" ".join(lexicon.values()).split(" "))
            del phoneme_counts[" "]
            return lexicon, phoneme_counts
        
        return lexicon 

    def extract_features(self, corpus):

        # get phonemes
        lexicon, phoneme_counts = self.get_phonemes(corpus)

        # TODO: Think of a more elegant way of doing this...
        #       The phonemic diversity calculation doesn't follow
        #       the same flow as other diversities because we
        #       precomputed the Z matrix and the main thing 
        #       needed now is the phoneme_counts.
        return phoneme_counts, phoneme_counts

    def calculate_similarities(self, features):

        unwanted = set(features) - set(self.Z)
        for unwanted_key in unwanted: del features[unwanted_key]
        phoneme_list = list(features.keys())
        Z = self.Z.loc[phoneme_list][phoneme_list].to_numpy()

        return Z

    def calculate_abundance(self, species):
        p = np.array(list(species.values()), dtype=np.float16)
        p /= p.sum()
        return p

    def __call__(self, corpus): 
        return super().__call__(corpus)


if __name__ == '__main__':

    # TEST
    lo_div = ['one massive earth', 'an enormous globe', 'the colossal world', '_']
    hi_div = ['basic human right', 'you were right', 'make a right', '_']

    # diversities
    print("diversities")
    print_div_metric(RhythmicDiversity, lo_div, hi_div)
    print_div_metric(PhonemicDiversity, lo_div, hi_div)

    # similarities
    print("similarities")
    print_sim_metric(RhythmicDiversity, lo_div, hi_div)
    print_sim_metric(PhonemicDiversity, lo_div, hi_div)


# TextDiversity pkgs
import os
import string
import itertools
import requests
from itertools import chain

import torch
import numpy as np
from sklearn.decomposition import PCA
from scipy.spatial import distance

import transformers
from transformers import AutoModel, AutoTokenizer
from sentence_transformers import SentenceTransformer

from gensim.utils import tokenize
import gensim.downloader

from nltk.corpus import stopwords

# locals
from ..metric import TextDiversity
from ..utils import *

# config
transformers.logging.set_verbosity_error()

# ==================================================================================
# helper functions 
# ==================================================================================

# NA

# ==================================================================================


class TokenSemanticDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'distance_fn': distance.chebyshev, 
        'dim_reducer': PCA,
        'remove_stopwords': False, 
        'remove_punct': False,
        'scale_dist': "exp", 
        'power_reg': False, 
        'mean_adj': True,
        'verbose': False,
        # TokenSemanticDiversity configs
        'MODEL_NAME': "bert-base-uncased", # "roberta-base", "microsoft/deberta-large", # "bert-base-uncased", # "facebook/bart-large-cnn",
        'batch_size': 16,
        'use_gpu': True,
        'n_components': None 
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)
        self.model = AutoModel.from_pretrained(config['MODEL_NAME'])
        self.tokenizer = AutoTokenizer.from_pretrained(config['MODEL_NAME'])
        self.undesirable_tokens = [
            self.tokenizer.pad_token_id, 
            self.tokenizer.cls_token_id, 
            self.tokenizer.sep_token_id
        ]
        self.batch_size = config['batch_size']
        self.device = torch.device('cuda' if config['use_gpu'] and torch.cuda.is_available() else 'cpu')

        # move model to device
        if isinstance(self.model, torch.nn.Module):
            self.model.to(self.device)

        # print('{0} ({1})'.format(self.__class__.__name__, config['MODEL_NAME']))

    def encode(self, input_ids, attention_mask):
        self.model.eval()
        with torch.no_grad():
            out = self.model(input_ids, attention_mask=attention_mask, output_hidden_states=True)
        # emb = out.last_hidden_state
        emb = out.hidden_states[-2]
        # emb = torch.stack(out.hidden_states[-4:]).mean(axis=0)
        # emb = torch.stack(out.hidden_states[-4:]).sum(axis=0)
        # emb = torch.cat(out.hidden_states[-4:], dim=-1)
        return emb

    def extract_features(self, corpus):
        inputs = self.tokenizer(corpus, return_tensors='pt', padding=True, truncation=True)
        batches = zip(chunker(inputs.input_ids, self.batch_size), 
                      chunker(inputs.attention_mask, self.batch_size))
        if self.config['verbose']:
            print('getting token embeddings...')
            batches = tqdm(batches, total=int(len(inputs.input_ids)/self.batch_size))

        outputs = []
        for input_ids, attention_mask in batches:
            emb = self.encode(input_ids.to(self.device), 
                       attention_mask.to(self.device))
            outputs.append(emb)
        embeddings = torch.cat(outputs)

        # remove undesirable tokens
        idx = np.isin(inputs['input_ids'],  self.undesirable_tokens, assume_unique=True, invert=True).reshape(-1)
        # idx = np.isin(inputs['input_ids'],  [self.tokenizer.cls_token_id], assume_unique=True).reshape(-1)
        tok = np.array(self.tokenizer.convert_ids_to_tokens(inputs.input_ids.view(-1)))[idx]
        boe = embeddings.view(-1, embeddings.shape[-1])[idx].detach().cpu().numpy()

        # remove stopwords
        if self.config['remove_stopwords']:
            idx = np.isin(tok, stopwords.words('english'), invert=True)
            tok = tok[idx]
            boe = boe[idx]

        # remove punctuation
        if self.config['remove_punct']:
            punct = '''!()-[]{};:'"\, <>./?@#$%^&*_~'''
            punct = [c for c in punct]
            idx = np.isin(tok, punct, invert=True)
            tok = tok[idx]
            boe = boe[idx]

        # compress embedding to speed up similarity matrix computation
        if self.config['n_components'] == "auto":
            n_components = min(max(2, len(boe) // 10), boe.shape[-1])
            if self.config['verbose']:
                print('Using n_components={}'.format(str(n_components)))
        else:
            n_components = -1

        if type(n_components) == int and n_components > 0 and len(boe) > 1:
            boe = self.config['dim_reducer'](n_components=n_components).fit_transform(boe)

        if len(np.flatnonzero(np.core.defchararray.find(tok,'##')!=-1)) > 0:
            tok, boe = merge_bpe(tok, boe)

        return boe, tok

    # def calculate_similarities(self, features):
    #     Z = compute_pairwise(features, self.config['distance_fn'])
    #     Z = Z - Z.mean()
    #     Z = 1 / (1+np.e**Z)
    #     np.fill_diagonal(Z, 1)
    #     return Z

    def calculate_similarities(self, features):
        num_embeddings = len(features)

        Z = np.eye(num_embeddings)
        iu = np.triu_indices(num_embeddings, k=1)
        il = (iu[1], iu[0])

        iterable = range(num_embeddings)
        if self.config['verbose']:
            print('calculating similarity matrix...')
            iterable = tqdm(iterable)

        for e1 in iterable:
            for e2 in range(1, num_embeddings - e1):
                d = self.config['distance_fn'](features[e1], features[e1 + e2])
                if self.config['scale_dist'] == "exp":
                    d = np.exp(-d)
                elif self.config['scale_dist'] == "invert":
                    d = 1 - d
                Z[e1][e1 + e2] = d     
        Z[il] = Z[iu]

        # remove some noise from the Z similarities
        if self.config['power_reg']:
            Z **= 2

        # remove some noise from the Z similarities
        if self.config['mean_adj']:
            off_diag = np.where(~np.eye(Z.shape[0],dtype=bool))
            Z[off_diag] -= Z[off_diag].mean()
            Z = np.where(Z < 0, 0 , Z)

        return Z

    def calculate_abundance(self, species):
        num_species = len(species)
        p = np.full(num_species, 1 / num_species)
        return p

    def __call__(self, response_set): 
        return super().__call__(response_set)


class TokenEmbeddingDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'distance_fn': distance.chebyshev, 
        'dim_reducer': PCA,
        'remove_stopwords': False, 
        'remove_punct': False,
        'scale_dist': "exp", 
        'power_reg': False, 
        'mean_adj': True,
        'verbose': False,
        # TokenEmbeddingDiversity configs
        'EMBEDDING': 'word2vec-google-news-300', # glove-wiki-gigaword-300, fasttext-wiki-news-subwords-300, 'word2vec-google-news-300'
        'batch_size': 16,
        'n_components': None  
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)

        self.model = gensim.downloader.load(config['EMBEDDING'])
        self.batch_size = config['batch_size']

        # print('{0} ({1})'.format(self.__class__.__name__, config['EMBEDDING']))


    def extract_features(self, corpus):

        # tokenize
        tok = []
        for s in corpus:
            tok.append(list(tokenize(s, lowercase=True)))
        tok = list(itertools.chain(*tok))
        tok = [t for t in tok if t in self.model.index_to_key]

        # embed
        embeddings = [self.model[t] for t in tok]

        tok = np.array(tok)
        boe = np.stack(embeddings)

        # remove stopwords
        if self.config['remove_stopwords']:
            idx = np.isin(tok, stopwords.words('english'), invert=True)
            tok = tok[idx]
            boe = boe[idx]

        # remove punctuation
        if self.config['remove_punct']:
            punct = '''!()-[]{};:'"\, <>./?@#$%^&*_~'''
            punct = [c for c in punct]
            idx = np.isin(tok, punct, invert=True)
            tok = tok[idx]
            boe = boe[idx]

        # compress embedding to speed up similarity matrix computation
        if self.config['n_components'] == "auto":
            n_components = min(max(2, len(boe) // 10), boe.shape[-1])
            if self.config['verbose']:
                print('Using n_components={}'.format(str(n_components)))
        else:
            n_components = -1

        if type(n_components) == int and n_components > 0 and len(boe) > 1:
            boe = self.config['dim_reducer'](n_components=n_components).fit_transform(boe)

        if len(np.flatnonzero(np.core.defchararray.find(tok,'##')!=-1)) > 0:
            tok, boe = merge_bpe(tok, boe)

        return boe, tok

    def calculate_similarities(self, features):
        num_embeddings = len(features)

        Z = np.eye(num_embeddings)
        iu = np.triu_indices(num_embeddings, k=1)
        il = (iu[1], iu[0])

        iterable = range(num_embeddings)
        if self.config['verbose']:
            print('calculating similarity matrix...')
            iterable = tqdm(iterable)

        for e1 in iterable:
            for e2 in range(1, num_embeddings - e1):
                d = self.config['distance_fn'](features[e1], features[e1 + e2])
                if self.config['scale_dist'] == "exp":
                    d = np.exp(-d)
                elif self.config['scale_dist'] == "invert":
                    d = 1 - d
                Z[e1][e1 + e2] = d     
        Z[il] = Z[iu]

        # remove some noise from the Z similarities
        if self.config['power_reg']:
            Z **= 2

        # remove some noise from the Z similarities
        if self.config['mean_adj']:
            off_diag = np.where(~np.eye(Z.shape[0],dtype=bool))
            Z[off_diag] -= Z[off_diag].mean()
            Z = np.where(Z < 0, 0 , Z)

        return Z

    def calculate_abundance(self, species):
        num_species = len(species)
        p = np.full(num_species, 1 / num_species)
        return p

    def __call__(self, response_set): 
        return super().__call__(response_set)


class STokenSemanticDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'distance_fn': distance.cosine, 
        'dim_reducer': PCA,
        'remove_stopwords': False, 
        'remove_punct': True,
        'scale_dist': "invert", 
        'power_reg': False, 
        'mean_adj': False,
        'verbose': False,
        # SentenceSemanticDiversity configs
        'MODEL_NAME':"bert-large-nli-stsb-mean-tokens",
        'use_gpu': True,
        'n_components': None 
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)
        self.device = torch.device('cuda' if config['use_gpu'] and torch.cuda.is_available() else 'cpu')
        self.model = SentenceTransformer(config['MODEL_NAME'], device=self.device)
        self.config['verbose'] = config['verbose']
        self.undesirable_tokens = [
            self.model.tokenizer.pad_token_id, 
            self.model.tokenizer.cls_token_id, 
            self.model.tokenizer.sep_token_id
        ]

        # print('{0} ({1})'.format(self.__class__.__name__, config['MODEL_NAME']))

    def extract_features(self, corpus):

        inputs = self.model.tokenizer(corpus)
        tok = list(chain.from_iterable(inputs.input_ids))

        embeddings = self.model.encode(corpus, output_value='token_embeddings')

        boe = torch.cat(embeddings).numpy()   

        # remove undesirable tokens
        idx = np.isin(tok, self.undesirable_tokens, assume_unique=True, invert=True).reshape(-1)
        # idx = np.isin(inputs['input_ids'],  [self.tokenizer.cls_token_id], assume_unique=True).reshape(-1)
        tok = np.array(self.model.tokenizer.convert_ids_to_tokens(torch.tensor(tok).view(-1)))

        assert len(boe) == len(tok), "boe.shape: {}\n tok.shape: {}".format(boe.shape, tok.shape)

        tok = tok[idx]
        boe = boe[idx]

        # remove stopwords
        if self.config['remove_stopwords']:
            idx = np.isin(tok, stopwords.words('english'), invert=True)
            tok = tok[idx]
            boe = boe[idx]

        # remove punctuation
        if self.config['remove_punct']:
            punct = '''!()-[]{};:'"\, <>./?@#$%^&*_~'''
            punct = [c for c in punct]
            idx = np.isin(tok, punct, invert=True)
            tok = tok[idx]
            boe = boe[idx]
        
        # compress embedding to speed up similarity matrix computation
        if self.config['n_components'] == "auto":
            n_components = min(max(2, len(boe) // 10), boe.shape[-1])
            if self.config['verbose']:
                print('Using n_components={}'.format(str(n_components)))
        else:
            n_components = -1

        if type(n_components) == int and n_components > 0 and len(boe) > 1:
            boe = self.config['dim_reducer'](n_components=n_components).fit_transform(boe)

        if len(np.flatnonzero(np.core.defchararray.find(tok,'##')!=-1)) > 0:
            tok, boe = merge_bpe(tok, boe)

        return boe, tok

    def calculate_similarities(self, features):
        num_embeddings = len(features)

        Z = np.eye(num_embeddings)
        iu = np.triu_indices(num_embeddings, k=1)
        il = (iu[1], iu[0])

        iterable = range(num_embeddings)
        if self.config['verbose']:
            print('calculating similarity matrix...')
            iterable = tqdm(iterable)

        for e1 in iterable:
            for e2 in range(1, num_embeddings - e1):
                d = self.config['distance_fn'](features[e1], features[e1 + e2])
                if self.config['scale_dist'] == "exp":
                    d = np.exp(-d)
                elif self.config['scale_dist'] == "invert":
                    d = 1 - d
                Z[e1][e1 + e2] = d     
        Z[il] = Z[iu]

        # remove some noise from the Z similarities
        if self.config['power_reg']:
            Z **= 2

        # remove some noise from the Z similarities
        if self.config['mean_adj']:
            off_diag = np.where(~np.eye(Z.shape[0],dtype=bool))
            Z[off_diag] -= Z[off_diag].mean()
            Z = np.where(Z < 0, 0 , Z)

        return Z

    def calculate_abundance(self, species):
        num_species = len(species)
        p = np.full(num_species, 1 / num_species)
        return p

    def __call__(self, response_set): 
        return super().__call__(response_set)


class DocumentSemanticDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'distance_fn': distance.cosine, 
        'dim_reducer': PCA,
        'remove_stopwords': False, 
        'scale_dist': "invert", 
        'power_reg': False, 
        'mean_adj': False,
        'verbose': False,
        # DocumentSemanticDiversity configs
        'MODEL_NAME': "princeton-nlp/sup-simcse-roberta-large", # "bert-large-nli-stsb-mean-tokens",
        'use_gpu': True,
        'n_components': None
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)
        self.device = torch.device('cuda' if config['use_gpu'] and torch.cuda.is_available() else 'cpu')
        self.model = SentenceTransformer(config['MODEL_NAME'], device=self.device)
        self.config['verbose'] = config['verbose']

        # print('{0} ({1})'.format(self.__class__.__name__, config['MODEL_NAME']))

    def extract_features(self, corpus):

        boe = np.stack(self.model.encode(corpus))
        
        # compress embedding to speed up similarity matrix computation
        if self.config['n_components'] == "auto":
            n_components = min(max(2, len(boe) // 10), boe.shape[-1])
            if self.config['verbose']:
                print('Using n_components={}'.format(str(n_components)))
        else:
            n_components = -1

        if type(n_components) == int and n_components > 0 and len(boe) > 1:
            boe = self.config['dim_reducer'](n_components=n_components).fit_transform(boe)

        return boe, corpus

    def calculate_similarities(self, features):
        num_embeddings = len(features)

        Z = np.eye(num_embeddings)
        iu = np.triu_indices(num_embeddings, k=1)
        il = (iu[1], iu[0])

        iterable = range(num_embeddings)
        if self.config['verbose']:
            print('calculating similarity matrix...')
            iterable = tqdm(iterable)

        for e1 in iterable:
            for e2 in range(1, num_embeddings - e1):
                d = self.config['distance_fn'](features[e1], features[e1 + e2])
                if self.config['scale_dist'] == "exp":
                    d = np.exp(-d)
                elif self.config['scale_dist'] == "invert":
                    d = 1 - d
                Z[e1][e1 + e2] = d     
        Z[il] = Z[iu]

        # remove some noise from the Z similarities
        if self.config['power_reg']:
            Z **= 2

        # remove some noise from the Z similarities
        if self.config['mean_adj']:
            off_diag = np.where(~np.eye(Z.shape[0],dtype=bool))
            Z[off_diag] -= Z[off_diag].mean()
            Z = np.where(Z < 0, 0 , Z)

        return Z

    def calculate_abundance(self, species):
        num_species = len(species)
        p = np.full(num_species, 1 / num_species)
        return p

    def __call__(self, response_set): 
        return super().__call__(response_set)


if __name__ == '__main__':

    # TEST
    lo_div = ['one massive earth', 'an enormous globe', 'the colossal world']
    hi_div = ['basic human right', 'you were right', 'make a right']

    # diversities
    print("diversities")
    print_div_metric(TokenSemanticDiversity, lo_div, hi_div)
    print_div_metric(TokenEmbeddingDiversity, lo_div, hi_div)
    print_div_metric(STokenSemanticDiversity, lo_div, hi_div)
    print_div_metric(DocumentSemanticDiversity, lo_div, hi_div)

    # similarities
    print("similarities")
    print_sim_metric(TokenSemanticDiversity, lo_div, hi_div)
    print_sim_metric(TokenEmbeddingDiversity, lo_div, hi_div)
    print_sim_metric(STokenSemanticDiversity, lo_div, hi_div)
    print_sim_metric(DocumentSemanticDiversity, lo_div, hi_div)


import os
import requests
import re
from tqdm import tqdm
import zipfile
import numpy as np
import torch
from spacy.lang.en import English

def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))

def merge_bpe(tok, boe, chars="##"):
    new_tok = []
    new_boe = []

    emb = []
    append = ""
    for t, e in zip(tok[::-1], boe[::-1]):
        t += append
        emb.append(e)
        if t.startswith(chars):
            append = t.replace(chars, "")
        else:
            append = ""
            new_tok.append(t)
            new_boe.append(np.stack(emb).mean(axis=0))
            emb = []  
    new_tok = np.array(new_tok)[::-1]
    new_boe = np.array(new_boe)[::-1]
    
    return new_tok, new_boe

def find_max_list(lists):
    list_len = [len(l) for l in lists]
    return max(list_len)

def cos_sim(a, b):
    """
    Computes the cosine similarity cos_sim(a[i], b[j]) for all i and j.
    :return: Matrix with res[i][j]  = cos_sim(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    a_norm = torch.nn.functional.normalize(a, p=2, dim=1)
    b_norm = torch.nn.functional.normalize(b, p=2, dim=1)
    return torch.mm(a_norm, b_norm.transpose(0, 1))

def compute_pairwise(inputs, fn, diagonal_val=0, verbose=True):
    num_inputs = len(inputs)

    Z = np.eye(num_inputs)
    iu = np.triu_indices(num_inputs, k=1)
    il = (iu[1], iu[0])

    outer_loop = range(num_inputs)
    if verbose:
        outer_loop = tqdm(outer_loop)

    for i in outer_loop:
        for j in range(1, num_inputs - i):
            d = fn(inputs[i], inputs[i + j])
            Z[i][i + j] = d     
    Z[il] = Z[iu]

    if diagonal_val is not None:
        np.fill_diagonal(Z, diagonal_val)
        
    return Z

def clean_text(texts):
    texts = [text.replace("<br />", "") for text in texts]
    texts = [text.replace("...", ". ") for text in texts]
    texts = [text.replace("..", ". ") for text in texts]
    texts = [text.replace(".", ". ") for text in texts]
    texts = [text.replace("!", "! ") for text in texts]
    texts = [text.replace("?", "? ") for text in texts]
    texts = [text.strip() for text in texts]
    return texts

def split_sentences(texts):
    sentencizer = English()
    sentencizer.add_pipe("sentencizer")

    sentences = []
    for text in texts:
        sents = list(sentencizer(text).sents)
        sents = [s.text.strip() for s in sents if s.text.strip()]
        sentences.extend(sents)
      
    return sentences

def print_div_metric(metric_class, lo_div, hi_div):

    div_metric = metric_class({'normalize': False})
    div_lo_unnorm = div_metric(lo_div)
    div_hi_unnorm = div_metric(hi_div)

    div_metric.config['normalize'] = True
    div_lo_norm = div_metric(lo_div)
    div_hi_norm = div_metric(hi_div)
    
    print('metric:{0} | lo_div: {1:0.3f} ({2:0.3f}) | hi_div: {3:0.3f} ({4:0.3f}) | passed: {5}'
        .format(
            metric_class.__name__,
            div_lo_unnorm, 
            div_lo_norm, 
            div_hi_unnorm,
            div_hi_norm,
            div_lo_unnorm < div_hi_unnorm 
        and div_lo_norm < div_hi_norm))

def print_sim_metric(metric_class, lo_div, hi_div):

    div_metric = metric_class()
    sim_hi = div_metric.similarity(lo_div)
    sim_lo = div_metric.similarity(hi_div)

    print('metric:{0} | lo_div: {1:0.3f} | hi_div: {2:0.3f}'
        .format(
            metric_class.__name__,
            sim_lo, 
            sim_hi))

if __name__ == '__main__':
    pass


# TextDiversity pkgs
import spacy
import torch
import numpy as np
from functools import partial
from sklearn.decomposition import PCA

# tree libs
import networkx as nx
from karateclub import FeatherGraph, GL2Vec, Graph2Vec, LDP
import zss

# # graphviz has compatibility issues on windows
# from networkx.drawing.nx_pydot import graphviz_layout
# from matplotlib import pyplot as plt

# locals
from ..metric import TextDiversity
from ..utils import *

# ==================================================================================
# helper functions 
# ==================================================================================

# def plot_graph(G):
#     pos = nx.nx_agraph.pygraphviz_layout(G, prog="dot")
#     num_nodes = G.number_of_nodes()
#     # fig = plt.figure(1, figsize=(max(12, num_nodes), max(6, num_nodes/2)), dpi=60)
#     nx.draw_networkx(G, pos, with_labels=False, node_color='#00b4d9')
#     # draw node labels
#     node_labels = nx.get_node_attributes(G, 'text') 
#     nx.draw_networkx_labels(G, pos, node_labels)
#     # draw edge labels
#     edge_labels = nx.get_edge_attributes(G,'dep')
#     nx.draw_networkx_edge_labels(G, pos, edge_labels)
#     plt.show()

# used with nx.graph_edit_distance() ===============================================
def node_match_on_pos(G1_node, G2_node):
    return G1_node['pos'] == G2_node['pos']

def edge_match_on_dep(G1_edge, G2_edge):
    return G1_edge['dep'] == G2_edge['dep']

# used with zss ====================================================================
def get_nodes_dict(T):
    nodes_dict = {}
    for edge in T.edges():
        if edge[0] not in nodes_dict:
            nodes_dict[edge[0]] = zss.Node(edge[0])
        if edge[1] not in nodes_dict:
            nodes_dict[edge[1]] = zss.Node(edge[1])
        nodes_dict[edge[0]].addkid(nodes_dict[edge[1]])
    return nodes_dict

def zss_tree_edit_distance(G1, G2):
    source1 = [n for (n, d) in G1.in_degree() if d == 0][0]
    T1 = nx.dfs_tree(G1, source=source1)
    T1_nodes_dict = get_nodes_dict(T1)
    
    source2 = [n for (n, d) in G2.in_degree() if d == 0][0]
    T2 = nx.dfs_tree(G2, source=source2)
    T2_nodes_dict = get_nodes_dict(T2)

    return zss.simple_distance(T1_nodes_dict[source1], T2_nodes_dict[source2])

# ==================================================================================

class DependencyDiversity(TextDiversity):

    default_config = {
        # TextDiversity configs
        'q': 1,
        'normalize': False,
        'dim_reducer': PCA,
        'remove_stopwords': False, 
        'verbose': False,
        # DependencyDiversity configs
        'similarity_type':"ldp",
        'use_gpu': False,
        'n_components': None 
    }

    def __init__(self, config={}):
        config = {**self.default_config, **config} 
        super().__init__(config)
        self.model = spacy.load("en_core_web_sm")

    def generate_dependency_tree(self, in_string):
        ''' 
        NOTES: 
          - Use the index instead of the token to avoid loops
          - Ensure the node is a string since zss tree edit distance requires it
        '''
        doc = self.model(in_string)

        G = nx.DiGraph()

        nodes = [(str(token.i), {'text': token.text, 'pos' : token.pos_}) 
                 for token in doc 
                 # if not token.is_punct
        ]
        G.add_nodes_from(nodes)

        edges = [(str(token.head.i), str(token.i), {'dep' : token.dep_}) 
                 for token in doc 
                 if token.head.i != token.i 
                 # and not token.is_punct
                 # and not token.head.is_punct
        ]
        G.add_edges_from(edges)

        return G

    def extract_features(self, corpus):

        # clean corpus
        corpus = clean_text(corpus)

        # split sentences
        sentences = split_sentences(corpus)

        # generate dependency tree graphs
        graphs = [self.generate_dependency_tree(s) for s in sentences]

        return graphs, sentences

    def calculate_similarities(self, features):
        """
        self.similarity_type: 
          - "graph_edit_distance" --> nx.graph_edit_distance
          - "tree_edit_distance" --> zss.simple_distance
          - "ldp" --> karateclub.LDP (Local Degree Profile)
          - "feather" --> karateclub.FeatherGraph
        """

        if 'distance' in self.config['similarity_type']:

            if self.config['similarity_type'] == "graph_edit_distance":
                Z = compute_pairwise(features, partial(nx.graph_edit_distance, 
                                                        node_match=node_match_on_pos, 
                                                        edge_match=edge_match_on_dep))
            elif self.config['similarity_type'] == "tree_edit_distance":
                Z = compute_pairwise(features, zss_tree_edit_distance)

            # convert distance to similarity

            # # option 1
            # Z = 1 - (Z / (Z.max()))
            # Z **= 4
            # np.fill_diagonal(Z, 1)

            # option 2
            Z = Z - Z.mean()
            Z = 1 / (1+np.e**Z)
            np.fill_diagonal(Z, 1)

        else:

            # the embedding approaches require integer node labels
            features = [nx.convert_node_labels_to_integers(g, first_label=0, ordering='default') for g in features]

            if self.config['similarity_type'] == "ldp":
                model = LDP(bins=64) # more bins, less similarity
                model.fit(features)
                emb = model.get_embedding().astype(np.float32)
            elif self.config['similarity_type'] == "feather":
                model = FeatherGraph(theta_max=100) # higher theta, less similarity
                model.fit(features)
                emb = model.get_embedding().astype(np.float32)

            # compress embedding to speed up similarity matrix computation
            if self.config['n_components'] == "auto":
                n_components = min(max(2, len(emb) // 10), emb.shape[-1])
                if self.config['verbose']:
                    print('Using n_components={}'.format(str(n_components)))
            else:
                n_components = -1

            if type(n_components) == int and n_components > 0 and len(emb) > 1:
                emb = self.config['dim_reducer'](n_components=n_components).fit_transform(emb)

            Z = cos_sim(emb, emb).numpy()

            # strongly penalize for any differences to make Z more intuitive
            Z **= 200

        return Z

    def calculate_abundance(self, species):
        num_species = len(species)
        p = np.full(num_species, 1 / num_species)
        return p

    def __call__(self, corpus): 
        return super().__call__(corpus)


if __name__ == '__main__':

    # TEST
    lo_div = ['one massive earth', 'an enormous globe', 'the colossal world']
    hi_div = ['basic human right', 'you were right', 'make a right']

    # diversities
    print("diversities")
    print_div_metric(DependencyDiversity, lo_div, hi_div)

    # similarities
    print("similarities")
    print_sim_metric(DependencyDiversity, lo_div, hi_div)
